local MAJOR_VERSION = "Threat-1.0"
local MINOR_VERSION = tonumber(("$Revision: 37586 $"):match("%d+"))

if _G.ThreatLib_OUT_OF_DATE or not AceLibrary:IsNewVersion(MAJOR_VERSION, MINOR_VERSION) then
	_G.ThreatLib_OUT_OF_DATE = true
	return
end

if MINOR_VERSION > _G.ThreatLib_MINOR_VERSION then
	_G.ThreatLib_MINOR_VERSION = MINOR_VERSION
end

local ThreatLib = _G.ThreatLib

local BB = AceLibrary("Babble-Boss-2.2")

-- You need to register the module with the localized mob name, as this is used to 
local c = BB["Hydross the Unstable"]

local ThreatLibNPCModuleCore = ThreatLib:GetModule("NPCCore")
ThreatLibNPCModuleCore:SetModuleDefaultState(c, false)
local Hydross = ThreatLibNPCModuleCore:NewModule(c)

local poisonTriggerStrings = {
	["enUS"] = "Aaghh, the poison...",
	["deDE"] = "Aahh, das Gift...",
	["frFR"] = "Aaarrgh, le poison…",
	["koKR"] = "으아아, 독이...",
	["zhTW"] = nil
}

local waterTriggerStrings = {
	["enUS"] = "Better, much better.",
	["deDE"] = "Besser, viel besser.",
	["frFR"] = "Ça va mieux. Beaucoup mieux.",
	["koKR"] = "아... 기분이 훨씬 좋군.",
	["zhTW"] = nil
}

local spawnNameStrings = {
	["Tainted"] = {
		["enUS"] = "Tainted Spawn of Hydross",
		["deDE"] = nil,
		["frFR"] = nil,
		["koKR"] = nil,
		["zhTW"] = nil
	},
	["Pure"] = {
		["enUS"] = "Pure Spawn of Hydross",
		["deDE"] = nil,
		["frFR"] = nil,
		["koKR"] = nil,
		["zhTW"] = nil
	}
}

function Hydross:Init()
	local poisonPhase = triggerStrings[_G.GetLocale()]
	local waterPhase = triggerStrings[_G.GetLocale()]
	poisonTriggerStrings = nil
	waterTriggerStrings = nil
	
	self:RegisterCombatant(BB["Hydross the Unstable"], true)
	
	local pureSpawnName = spawnNameStrings["Tainted"][_G.GetLocale()]
	local taintedSpawnName = spawnNameStrings["Pure"][_G.GetLocale()]
	if pureSpawnName and taintedSpawnName then
		self:RegisterCombatant(pureSpawnName, self.spawnDeath)
		self:RegisterCombatant(taintedSpawnName, self.spawnDeath)
	end

	-- To register a callback to a yell or emote:
	-- self:RegisterChatEvent([yell|event], localizedMobName, yellOrEmoteToTriggerOn, callbackFunction)
	
	self:RegisterChatEvent("yell", BB["Hydross the Unstable"], poisonPhase, self.phaseTransition)
	self:RegisterChatEvent("yell", BB["Hydross the Unstable"], waterPhase, self.phaseTransition)
end

function Hydross:StartFight()
	self:SetNumberOfMobs(1)
end

function Hydross:spawnDeath()
	self:SetNumberOfMobs(self:GetNumberOfMobs() - 1)
end

function Hydross:phaseTransition()
	self:SetNumberOfMobs(5)		-- Hydross + 4 spawns
	self:WipeRaidThreat("Hydross")
end
