local _G = _G
local tonumber = _G.tonumber
local error = _G.error
local pairs = _G.pairs
local select = _G.select
local next = _G.next
local tinsert = _G.tinsert
local type = _G.type
local time = _G.time
local sort = _G.table.sort
local tostring = _G.tostring
local setmetatable = _G.setmetatable

local UnitName = _G.UnitName
local GetNumRaidMembers = _G.GetNumRaidMembers
local GetNumPartyMembers = _G.GetNumPartyMembers
local InCombatLockdown = _G.InCombatLockdown
local UnitAffectingCombat = _G.UnitAffectingCombat
local GetTime = _G.GetTime
local UnitClass = _G.UnitClass
local ChatThrottleLib = _G.ChatThrottleLib

local AceLibrary = _G.AceLibrary

local MAJOR_VERSION = "Threat-1.0"
local MINOR_VERSION = tonumber(("$Revision: 38197 $"):match("%d+"))

_G.ThreatLib_OUT_OF_DATE = nil

if not AceLibrary then error(MAJOR_VERSION .. " requires AceLibrary") end
if not AceLibrary:IsNewVersion(MAJOR_VERSION, MINOR_VERSION) then
	_G.ThreatLib_OUT_OF_DATE = true
	return
end

local BS = AceLibrary("Babble-Spell-2.2")
_G.ThreatLib = AceLibrary("AceAddon-2.0"):new("AceConsole-2.0", "AceComm-2.0", "AceEvent-2.0", "AceModuleCore-2.0")
local ThreatLib = _G.ThreatLib
_G.ThreatLib_MINOR_VERSION = MINOR_VERSION
ThreatLib.OnCommReceive = {}

local playerName = UnitName("player")

local function NumericCheckSum(text)
	local counter = 1
	local len = text:len()
	for i = 1, len, 3 do
		counter = (counter*8257 % 16777259) +
			(text:byte(i)) +
			((text:byte(i+1) or 1)*127) +
			((text:byte(i+2) or 2)*16383)
	end
	return counter % 16777213
end

local HexCheckSum = setmetatable({}, {__mode='k', __index = function(self, key)
	if type(key) ~= "string" then -- REMOVE LATER MAYBE
		error(("Bad argument #1 to `HexCheckSum'. %q expected, got %q (%s). Please report to Antiarc."):format("string", type(key), tostring(key)), 2)
	end
	local value = ("%06x"):format(NumericCheckSum(key))
	self[key] = value
	return value
end})

local GLOBAL_HASH = HexCheckSum["_GLOBAL_"]
local mobNameTranslations = {}
local partyMemberLocales = {}
local partyMemberLocaleCount = 0
local translationHashes = {}

local new, newHash, newSet, del
do
	local list = setmetatable({}, {__mode='k'})
	function new(...)
		local t = next(list)
		if t then
			list[t] = nil
			for i = 1, select('#', ...) do
				t[i] = select(i, ...)
			end
			return t
		else
			return {...}
		end
	end
	ThreatLib.new = new
	function newHash(...)
		local t = next(list)
		if t then
			list[t] = nil
		else
			t = {}
		end	
		for i = 1, select('#', ...), 2 do
			t[select(i, ...)] = select(i+1, ...)
		end
		return t
	end
	ThreatLib.newHash = newHash
	function newSet(...)
		local t = next(list)
		if t then
			list[t] = nil
		else
			t = {}
		end	
		for i = 1, select('#', ...) do
			t[select(i, ...)] = true
		end
		return t
	end
	ThreatLib.newSet = newSet
	function del(t)
		setmetatable(t, nil)
		for k in pairs(t) do
			t[k] = nil
		end
		list[t] = true
		return nil
	end
	ThreatLib.del = del
end

ThreatLib.partyUnits = {}
local partyUnits = ThreatLib.partyUnits

function ThreatLib:OnInitialize()
	self:ToggleModuleActive("ClassCore", true)
	
	self.commPrefix = "Thr"
	
	self:RegisterEvent("Threat_ThreatUpdated_Internal", "ThreatUpdated")
	
	self:SetCommPrefix(self.commPrefix)
	self:RegisterMemoizations({
		"THREAT_UPDATE",
		"MOB_THREAT_CHANGE",
		"MISDIRECT_THREAT",
		"MOB_NAME_TRANSLATION_REQUEST",
		"MOB_NAME_TRANSLATION_REPLY",
		"WIPE_ALL_THREAT",
		"CLIENT_INFO",
		"REQUEST_CLIENT_INFO",
		"MOB_RAID_THREAT_MULTIPLE",
		"LEFT_COMBAT"
	})
	self:RegisterComm(self.commPrefix, "GROUP", "OnCommReceive")
	self:ThreatInitData()

	self:MobTranslationInit()
	self.latestSeenRevision = MINOR_VERSION
	
	self.GlobalTarget = "GLOBAL"
	
	self:SetUnitInCombat("player", UnitAffectingCombat("player"))	
	self:SetUnitInCombat("pet", UnitAffectingCombat("pet"))
	
	self:RegisterEvent("PARTY_MEMBERS_CHANGED")
	self:RegisterEvent("RAID_ROSTER_UPDATE", "PARTY_MEMBERS_CHANGED")
	self:RegisterEvent("UNIT_PET", "PARTY_MEMBERS_CHANGED")
	self:RegisterEvent("PLAYER_PET_CHANGED", "PARTY_MEMBERS_CHANGED")
	self:RegisterEvent("CHAT_MSG_ADDON")
	self:RegisterEvent("PARTY_MEMBER_DISABLE")
	self:RegisterEvent("AceEvent_FullyInitialized", "LoadedInit")
end

function ThreatLib:LoadedInit()
	self:SendCommMessage("GROUP", "REQUEST_CLIENT_INFO", MyLocale, MINOR_VERSION)
	self:PARTY_MEMBERS_CHANGED()
end

function ThreatLib:ThreatUpdatedForUnit(unitName, target, threat, fromPet)
	if not target then
		error("Nil target in ThreatUpdated - please inform Antiarc")
	end
	
	local pName = unitName
	if self:DoesPlayerHavePet() and fromPet then
		pName = self:PetThreat().petName
	end
			
	local pUnit = partyUnits[pName]
	
	-- Shouldn't need this
	-- target = target or "_GLOBAL_"

	local targetHash = HexCheckSum[target]

	local oldThreat = self:GetThreat(pName, target)
	self:_setThreat(pName, targetHash, threat)
	
	local gThreat = self:GetThreat(pName, GLOBAL_HASH)	
	if target == "_GLOBAL_" then
		self:SignalGlobalThreatUpdated(pName)
	else
		-- REMOVE ASSERTS LATER (May 31, 2007)
		assert(type(pName) == "string")
		assert(type(pUnit) == "string")
		assert(type(targetHash) == "string")
		assert(type(threat + gThreat) == "number")
		-- REMOVE ASSERTS LATER (May 31, 2007)
		self:TriggerEvent("Threat_ThreatUpdated", pName, pUnit, targetHash, threat + gThreat)
	end

	if fromPet then
		self:PublishPetThreat()
	else
		self:PublishThreat()
	end
end

function ThreatLib:ThreatUpdated(target, threat, fromPet)
	self:ThreatUpdatedForUnit(playerName, target, threat, fromPet)
end

local threatTargets = {}

function ThreatLib:SignalGlobalThreatUpdated(pName)	
	local pUnit = partyUnits[pName]
	if not pUnit then
		return
	end
	local data = threatTargets[pName]
	if not data then
		return
	end
	local gThreat = self:GetThreat(pName, GLOBAL_HASH)

	-- As people generate global threat, we'll want to see it on all targets, not just those that we've explicitly 
	-- generated threat on, so we'll send juts global threat with a single target "GLOBAL" that GUI authors can watch for.
	-- This is sent before the targets so that if a target is then sent that the GUI is interested in, it can override the GLOBAL target
	self:TriggerEvent("Threat_ThreatUpdated", pName, pUnit, self.GlobalTarget, gThreat)

	for k,v in pairs(data) do
		if k ~= GLOBAL_HASH and type(v) == "number" then
			-- REMOVE ASSERTS LATER (May 31, 2007)
			assert(type(pName) == "string")
			assert(type(pUnit) == "string")
			assert(type(k) == "string")
			assert(type(v + gThreat) == "number")
			-- REMOVE ASSERTS LATER (May 31, 2007)
			-- self:Print(string.format("Setting global threat for %s to %s (%s global, %s target)", k, gThreat + v, gThreat, v))
			self:TriggerEvent("Threat_ThreatUpdated", pName, pUnit, k, gThreat + v)
		end
	end	
end

local combatFlags = {}

function ThreatLib:SetUnitInCombat(unit, flag)
	if flag then
		combatFlags[unit] = time()
	else
		combatFlags[unit] = nil
	end
end

function ThreatLib:IsUnitInCombat(unit)
	return not not combatFlags[unit]
end

function ThreatLib:GetCombatTimeForUnit(unit)
	if combatFlags[unit] then
		return time() - combatFlags[unit]
	end
	return 0
end

-- Command handling
function ThreatLib.OnCommReceive:THREAT_UPDATE(prefix, sender, distribution, target, threat, petName)
	if not target or not threat then
		return
	end
	target = self:TranslateForeignMobHash(target)
	
	local pName = sender
	if petName and self:OwnsPet(sender, petName) then
		pName = petName
	elseif petName ~= nil then
		return	-- We were given a pet name for a pet we don't know about
	end
	self:_setThreat(pName, target, threat)
	if target == GLOBAL_HASH then
		self:SignalGlobalThreatUpdated(pName)
	else
		-- REMOVE ASSERTS LATER (May 31, 2007)
		assert(type(pName) == "string")
		assert(type(partyUnits[pName]) == "string")
		assert(type(target) == "string")
		assert(type(threat) == "number")
		-- REMOVE ASSERTS LATER (May 31, 2007)
		self:TriggerEvent("Threat_ThreatUpdated", pName, partyUnits[pName], target, threat)
	end
end

function ThreatLib.OnCommReceive:LEFT_COMBAT(prefix, sender, distribution, petName)
	if petName and self:OwnsPet(sender, petName) then
		self:_clearThreat(petName)
	else
		self:_clearThreat(sender)
	end
end

function ThreatLib.OnCommReceive:WIPE_ALL_THREAT(prefix, sender, distribution)
	local senderUnit = partyUnits[sender]
	if not senderUnit then
		return
	end
	if IsRaidLeader(senderUnit) or IsRaidOfficer(senderUnit) then
		self:ScheduleEvent(
			"ThreatLibThreatWipe",
			function(self)
				self:MyThreat():_reduceAllThreat(0)
				self:PetThreat():_reduceAllThreat(0)
			end,
			0.35, 
			self
		)	
	end
end

function ThreatLib.OnCommReceive:MOB_RAID_THREAT_WIPE(prefix, sender, distribution, mobName)
	mobName = self:TranslateForeignMobName(mobName)
	self:ScheduleEvent(
		"ThreatLibMobRaidThreatWipe-" .. mobName,
		function(self, m)
			self:MyThreat():_multiplyTargetThreat(0, mobName)
			self:PetThreat():_multiplyTargetThreat(0, mobName)
		end,
		0.5,
		self
	)	
end

function ThreatLib.OnCommReceive:MOB_RAID_THREAT_MULTIPLE(prefix, sender, distribution, mobName, multiplier)
	mobName = self:TranslateForeignMobName(mobName)
	self:ScheduleEvent(
		"ThreatLibMobRaidThreatChange-" .. mobName,
		function(self, m)
			self:MyThreat():_reduceAllThreat(m)
			self:PetThreat():_reduceAllThreat(m)
		end,
		0.5,
		self,
		multiplier
	)
end

function ThreatLib.OnCommReceive:MISDIRECT_THREAT(prefix, sender, distribution, targetPlayer, targetEnemy, amount)
	targetEnemy = self:TranslateForeignMobName(targetEnemy)
	if targetPlayer == playerName and select(2, UnitClass(partyUnits[sender])) == "HUNTER" then
		self:GetModule("ClassCore").activeModule:_addTargetThreat(amount, targetEnemy)
	end
end

function ThreatLib:RequestThreatClear()
	if IsRaidLeader("player") or IsRaidOfficer("player") then
		self:SendCommMessage("GROUP", "WIPE_ALL_THREAT")
	else
		self:Print("You must be a raid officer or the raid leader to do request a threat list clear.")
	end
end

function ThreatLib:UnitIsUnit(a, b)
	if type(a) ~= "string" then
		error(("Bad argument #2 to `UnitIsUnit'. expected %q, got %q (%s)"):format("string", type(a), tostring(a)), 2)
	end
	if type(b) ~= "string" then
		error(("Bad argument #3 to `UnitIsUnit'. expected %q, got %q (%s)"):format("string", type(b), tostring(b)), 2)
	end
	local tostring_b = tostring(b)
	if HexCheckSum[a] == tostring_b then
		return true
	end
	local tostring_a = tostring(a)
	return tostring_a == tostring_b or HexCheckSum[b] == tostring_a
end

function ThreatLib:_getTargetHash(target)
	return HexCheckSum[target]
end

function ThreatLib:_setThreat(sender, target, threat)
	local data = threatTargets[sender]
	if not data then
		data = new()
		threatTargets[sender] = data
	end	
	data[target] = threat
end

function ThreatLib:_clearThreat(player)
	player = player or playerName
	local data = threatTargets[player]
	if not data then
		return
	end
	data[GLOBAL_HASH] = 0

	assert(type(player) == "string")
	assert(type(partyUnits[player]) == "string")
	self:TriggerEvent("Threat_ThreatUpdated", player, partyUnits[player], self.GlobalTarget, 0)
	
	--[[
	for k, v in pairs(data) do
		if k ~= GLOBAL_HASH then
			self:_setThreat(player, k, o)
			-- REMOVE ASSERTS LATER (May 31, 2007)
			assert(type(k) == "string")
			-- REMOVE ASSERTS LATER (May 31, 2007)
			self:TriggerEvent("Threat_ThreatUpdated", player, partyUnits[player], k, 0)
		end
	end
	]]--
	threatTargets[player] = del(data)
end

function ThreatLib:GetThreat(player, target)
	if type(target) ~= "string" then
		error(("Bad argument #2 to `GetThreat'. expected %q, got %q (%s)"):format("string", type(target), tostring(target)), 2)
	end
	local data = threatTargets[player]
	if not data then
		return 0
	end
	local nThreat = data[target] or data[HexCheckSum[target]] or 0
	local gThreat = data[GLOBAL_HASH] or 0
	if target == GLOBAL_HASH or target == "_GLOBAL_" then
		return gThreat
	else
		return nThreat + gThreat
	end
end

-- Command sending
function ThreatLib:CancelThreatPublishPet(petName)
	if not self:IsEventScheduled("ThreatLibThreatPublishPet") then
		return
	end
	self:SendCommMessage("GROUP", "LEFT_COMBAT", petName)
	self:CancelScheduledEvent("ThreatLibThreatPublishPet")
end

function ThreatLib:CancelThreatPublish()
	if not self:IsEventScheduled("ThreatLibThreatPublish") then
		return
	end
	self:SendCommMessage("GROUP", "LEFT_COMBAT")
	self:CancelScheduledEvent("ThreatLibThreatPublish")
end

function ThreatLib:PARTY_MEMBER_DISABLE(arg1)
	self.OnCommReceive.LEFT_COMBAT(self, nil, arg1)
end

local lastSentThreat = {}

local function func(self)
	local activeModule = self:GetModule("ClassCore").activeModule
	if not activeModule then error("Error: PublishThreat() has been called without an active class module") end
	local globalThreat = activeModule.generalThreat or 0
	for k, v in pairs(activeModule.targetThreat) do
		local total = v + globalThreat
		if k and lastSentThreat[k] ~= total then
			lastSentThreat[k] = total
			self:SendCommMessage("GROUP", "THREAT_UPDATE", HexCheckSum[k], total)
		end
	end
end

function ThreatLib:PublishThreat()
	if self:IsEventScheduled("ThreatLibThreatPublish") then
		return
	end
	self:ScheduleRepeatingEvent("ThreatLibThreatPublish", func, self:GetPublishInterval(), self)	
end

local function petfunc(self)
	if self:DoesPlayerHavePet() then
		local activeModule = self:PetThreat()
		local globalThreat = activeModule.generalThreat or 0
		for k, v in pairs(activeModule.targetThreat) do
			local total = v + globalThreat
			if k and lastSentThreat[k] ~= total then
				lastSentThreat[k] = total
				self:SendCommMessage("GROUP", "THREAT_UPDATE", HexCheckSum[k], total, activeModule.petName)
			end
		end	
	end
end

function ThreatLib:PublishPetThreat()
	if self:IsEventScheduled("ThreatLibThreatPublishPet") then
		return
	end
	self:ScheduleRepeatingEvent("ThreatLibThreatPublishPet", petfunc, self:GetPublishInterval(), self)	
end

local currentPartySize = 0

function ThreatLib:GetPublishInterval()
	-- Scale publish interval from 0.5 to 2.0 based on party size
	-- We'll be at 0.5 sec for 0-5 party size, scale from 0.5 to 1.5 for 6-25 players, and stay at 2.0 for > 20 players
	-- This means that we'll transmit as much as 66% less data in a raid
	local pSize = currentPartySize
	local partyNum = max(0, pSize - 5)
	return min(1.5, 0.5 + (1 * (partyNum / 20)))
end

function ThreatLib:UnitInMeleeRange(unitID)
	return UnitExists(unitID) and UnitIsVisible(unitID) and IsSpellInRange(BS["Attack"], unitID)
end

function ThreatLib:MyThreat()
	return self:GetModule("ClassCore").activeModule
end

function ThreatLib:DoesPlayerHavePet()
	return self:GetModule("ClassCore").petModule ~= nil and self:GetModule("ClassCore").petModule.petName ~= nil
end

function ThreatLib:OwnsPet(owner, petName)
	return UnitName(partyUnits[owner] .. "pet") == petName
end

function ThreatLib:PetThreat()
	return self:GetModule("ClassCore").petModule
end

function ThreatLib:EncounterMobs()
	return max(1, self:GetModule("NPCCore").encounterMobCount or 1)
end

do
	local sortThreatDesc_values
	local function sortThreatDesc(alpha, bravo)
		return sortThreatDesc_values[alpha] > sortThreatDesc_values[bravo]
	end

	local function threatIter(t, key)
		local n = t.n + 1
		local k = t[n]
		if k == nil then
			del(t.values)
			del(t)
			return nil
		end
		t.n = n
		return k, t.values[k]
	end

	function ThreatLib:IterateGroupThreatForTarget(target)
		if type(target) ~= "string" then
			error(("Bad argument #2 to `IterateGroupThreatForTarget`. expected %q, got %q (%s)"):format("string", type(target), tostring(target)), 2)
		end
		local results = new()
		local tHash = HexCheckSum[target]
		local t = new()
		for k, v in pairs(threatTargets) do
			if v[target] then
				t[#t+1] = k
				results[k] = v[target]
			elseif v[tHash] then
				t[#t+1] = k
				results[k] = v[tHash]
			end
		end
		sortThreatDesc_values = results
		table_sort(t, sortThreatDesc)
		sortThreatDesc_values = nil
		t.values = results
		t.n = 0
		return threatIter, t, nil
	end
end

function ThreatLib:GetMaxThreatOnTarget(target)
	if type(target) ~= "string" then
		error(("Bad argument #2 to `GetMaxThreatOnTarget`. expected %q, got %q (%s)"):format("string", type(target), tostring(target)), 2)
	end
	local tHash = HexCheckSum[target]
	
	local max = 0
	local maxname = nil
	for k, v in pairs(threatTargets) do
		local gThreat = threatTargets[k][GLOBAL_HASH]
		local val = (v[target] or v[tHash]) + (gThreat or 0)
		if val and val > max then
			max = val
			maxname = k
		end
	end
	return max, maxname
end

function ThreatLib:SendThreatTo(targetPlayer, targetEnemy, threat)
	if type(targetPlayer) ~= "string" then
		error(("Bad argument #2 to `SendThreatTo`. expected %q, got %q (%s)"):format("string", type(targetPlayer), tostring(targetPlayer)), 2)
	end
	
	if self:DoesPlayerHavePet() and self:PetThreat().petName == targetPlayer then
		self:PetThreat():_addTargetThreat(threat, targetEnemy)
	else
		self:SendCommMessage("GROUP", "MISDIRECT_THREAT", targetPlayer, targetEnemy, threat)
	end
end

function ThreatLib:SetRaidThreatMultiplier(mob, amount)
	if not amount then
		return
	end
	-- Raw mob names needs to go in
	-- It will be localized automatically as necessary
	self:SendCommMessage("GROUP", "MOB_RAID_THREAT_MULTIPLE", mob, amount)
	self.OnCommReceive.MOB_RAID_THREAT_MULTIPLE(self, nil, nil, nil, mob, amount)
end

function ThreatLib:WipeRaidThreatOnMob(mob)
	self:SendCommMessage("GROUP", "MOB_RAID_THREAT_MULTIPLE", mob, amount)
end

local partyMemberRevisions = {}

--local cachedForeignThreatSources = {}
function ThreatLib:IsUsingForeignThreatSource(partyMember)
--	if cachedForeignThreatSources[partyMember] then
--		return cachedForeignThreatSources[partyMember]
--	end
	
	local partyUnit = partyUnits[partyMember]
	if not partyUnit then
		return true
	end
	partyUnit = partyUnit:gsub("pet", "")
	if partyUnit == "player" or partyUnit == "" then
		return false
	end
	
	local partyMemberName = UnitName(partyUnit)
	
	local r = not partyMemberRevisions[partyMemberName]
	
--	cachedForeignThreatSources[partyMember] = r
	return r
end

function ThreatLib:IsAnyPartyMemberUsingForeignThreatSource()
	local raid = GetNumRaidMembers()
	if raid > 0 then
		for i = 1, raid do
			local name = UnitName("raid" .. i)
			if not partyMemberRevisions[name] then
				return true
			end
		end
	else
		local party = GetNumPartyMembers()
		if party > 0 then
			for i = 1, party do
				local name = UnitName("party" .. i)
				if not partyMemberRevisions[name] then
					return true
				end
			end
		end
	end
	return false
end

function ThreatLib:PARTY_MEMBERS_CHANGED()
	for k in pairs(partyUnits) do
		partyUnits[k] = nil
	end
	local sizeBeforeUpdate = currentPartySize
	local numRaid = GetNumRaidMembers()
	if numRaid > 0 then
		currentPartySize = numRaid
		for i = 1, numRaid do
			local u = "raidpet" .. i
			local n = UnitName(u)
			if n then
				partyUnits[n] = u
			end
		end
		local n = UnitName("pet")
		if n then
			partyUnits[n] = "pet"
		end
		for i = 1, numRaid do
			local u = "raid" .. i
			partyUnits[UnitName(u)] = u
		end
	else
		local numParty = GetNumPartyMembers()
		if numParty > 0 then
			currentPartySize = numParty
			for i = 1, numParty do
				local u = "partypet" .. i
				local n = UnitName(u)
				if n then
					partyUnits[n] = u
				end
			end
			local n = UnitName("pet")
			if n then
				partyUnits[n] = "pet"
			end
			for i = 1, numParty do
				local u = "party" .. i
				partyUnits[UnitName(u)] = u
			end
		else
			currentPartySize = 0
			local n = UnitName("pet")
			if n then
				partyUnits[n] = "pet"
			end
		end
	end
	partyUnits[playerName] = "player"

	if currentPartySize > sizeBeforeUpdate and currentPartySize > 0 then
		-- Invalidate cached foreign threat indicators
		-- We don't do this on a per-name basis due to pets
		for k, v in pairs(cachedForeignThreatSources) do
			cachedForeignThreatSources[k] = nil
		end
		
		-- Let's be sociable and tell the newcomers about ourselves.
		self:PublishLocaleAndVersion()
	end
end

function ThreatLib:CHAT_MSG_ADDON(prefix, message, distribution, sender)

	-- ignore partymembers that are silly enough to run KLHTM and Threat-1.0
	-- This semi-legally uses the partymemberLocales table to achieve its goal.

	if sender == playerName or partyMemberLocales[sender] then
		return
	end
	if prefix == "KLHTM" and (distribution == "PARTY" or distribution == "RAID") then
		local command, data = message:match("^(%a+) ?(.*)$")
		if command then
			local func = self.OnKLHTMMessage[command]
			if func then
				func(self, sender, data)
			end
		end
	end
end

ThreatLib.OnKLHTMMessage = {}
function ThreatLib.OnKLHTMMessage:t(sender, data)
	local threat = tonumber(data)
	if not threat then
		return
	end
	-- FIXME: Use mob name auto-localization here
	if self.KLHTMtarget then
		self:ThreatUpdatedForUnit( sender, self.KLHTMtarget, threat)
	else
		self:ThreatUpdatedForUnit( sender, "_GLOBAL_", threat )
	end
end

function ThreatLib.OnKLHTMMessage:target(sender, data)
	if data and type(data) == "string" then
		self.KLHTMtarget = data
	end
end

function ThreatLib.OnKLHTMMessage:cleartarget(sender, data)
	self.KLHTMtarget = nil
end

function ThreatLib:SendKLHTMMessage(command, data)
	local distribution
	if GetNumRaidMembers() > 0 then
		distribution = "RAID"
	elseif GetNumPartyMembers() > 0 then
		distribution = "PARTY"
	else
		return
	end
	
	local message
	if data then
		message = command .. " " .. data
	else
		message = command
	end
	
	ChatThrottleLib:SendAddonMessage("NORMAL", "KLHTM", message, distribution)
end

local function func(revision, sender)
	ThreatLib:TriggerEvent("Threat_OutOfDateNotice", MINOR_VERSION, revision, sender)
end
function ThreatLib:CheckLatestRevision(revision, sender)
	if not revision then
		return -- needed for older clients
	end
	if revision > self.latestSeenRevision then
		self.outOfDate = true
		self.latestSeenRevision = revision
		self:ScheduleEvent("ThreatLib_NotifyOutOfDate", func, 0.5, revision, sender)
	end
end
------------------------------------------------------------------------------------
-- Mob name translations, for groups with clients of multiple locales
-- This automatically translates mob names across locales
------------------------------------------------------------------------------------

local MyLocale = _G.GetLocale()

function ThreatLib:MobTranslationInit()
	self.mobNameTranslationEnabled = false
end

function ThreatLib:PublishLocaleAndVersion()
	partyMemberLocales[UnitName("player")] = MyLocale
	self:SendCommMessage("GROUP", "CLIENT_INFO", MyLocale, MINOR_VERSION)
end

function ThreatLib.OnCommReceive:REQUEST_CLIENT_INFO(prefix, sender, distribution, locale, revision)
	partyMemberLocales[sender] = locale
	partyMemberRevisions[sender] = revision
	self:CheckLatestRevision(revision, sender)
	self:ScheduleEvent("ThreatLibUpdatePartyLocaleSettings", function(self) self:UpdatePartyLocaleSettings() end, 0.5, self)		
	self:PublishLocaleAndVersion()
end

function ThreatLib.OnCommReceive:CLIENT_INFO(prefix, sender, distribution, locale, revision)
	partyMemberLocales[sender] = locale
	partyMemberRevisions[sender] = revision
	self:CheckLatestRevision(revision, sender)
	self:ScheduleEvent("ThreatLibUpdatePartyLocaleSettings", function(self) self:UpdatePartyLocaleSettings() end, 0.5, self)		
end

function ThreatLib:UpdatePartyLocaleSettings()	
	local t = newHash()
	local ct = 0
	for k,v in pairs(partyMemberLocales) do
		if partyUnits[k] == nil then
			partyMemberLocales[k] = nil
		else
			if not t[v] then
				t[v] = true
				ct = ct + 1
			end
		end
	end
	del(t)
	partyMemberLocaleCount = ct
	
	if ct > 1 then
		self:EnableMobNameTranslation()
	else
		self:DisableMobNameTranslation()
	end
end

function ThreatLib:EnableMobNameTranslation()
	self.mobNameTranslationEnabled = true
	if not self:IsEventRegistered("UNIT_TARGET") then
		self:RegisterEvent("UNIT_TARGET")
	end
end

function ThreatLib:DisableMobNameTranslation()
	self.mobNameTranslationEnabled = false
	if self:IsEventRegistered("UNIT_TARGET") then
		self:UnregisterEvent("UNIT_TARGET")
	end
end

function ThreatLib.OnCommReceive:MOB_NAME_TRANSLATION_REQUEST(prefix, sender, distribution, requestTo, mobName, fromLocale)
	if requestTo ~= UnitName("player") then
		return -- This isn't meant for us
	end
	
	local targetName = UnitName("target")
	
	-- They're giving us the name for our target, so store it
	local t = mobNameTranslations[targetName]
	if not t then
		mobNameTranslations[targetName] = newHash(fromLocale, mobName)
		self:UpdateMobNameTranslationHashes()
	elseif not mobNameTranslations[targetName][fromLocale] then	
		mobNameTranslations[targetName][fromLocale] = mobName
		self:UpdateMobNameTranslationHashes()
	end
	self:SendCommMessage("GROUP", "MOB_NAME_TRANSLATION_REPLY", mobName, fromLocale, targetName, MyLocale)
end

function ThreatLib.OnCommReceive:MOB_NAME_TRANSLATION_REPLY(prefix, sender, distribution, lMobName, lLocale, fMobName, fLocale)
	-- We don't want translations to locales that aren't ours
	if lLocale ~= MyLocale and fLocale ~= MyLocale then
		return
	end 
	
	-- We don't really care who this is in reply to - we'll spy on all comms to build a list of target names

	local targetName = nil
	local targetLocale = nil
	local mobName = nil
	if fLocale == MyLocale then
		mobName = fMobName
		targetName = lMobName
		targetLocale = lLocale
	elseif lLocale == MyLocale then
		mobName = lMobName
		targetName = fMobName
		targetLocale = fLocale
	end
	
	-- Not going to worry about table recycling here, as we'll likely just keep mob names around for the entire session.
	-- Anyone think we should do otherwise? Clear on zone-out or something?
	local t = mobNameTranslations[mobName]
	if not t then
		mobNameTranslations[mobName] = {}
	end
	if not mobNameTranslations[mobName][targetLocale] then
		mobNameTranslations[mobName][targetLocale] = targetName
		self:UpdateMobNameTranslationHashes()
	end
end

-- Create a lookup table of foreign mob name hash = local mob name hash
function ThreatLib:UpdateMobNameTranslationHashes()
	for mobName, locales in pairs(mobNameTranslations) do
		local localMobHash = HexCheckSum[mobName]
		for locale, translation in pairs(locales) do
			if locale ~= MyLocale then
				translationHashes[HexCheckSum[translation]] = localMobHash
			end
		end
	end
end

function ThreatLib:TranslateForeignMobHash(mobHash)
	if not self.mobNameTranslationEnabled then
		return mobHash
	end
	return translationHashes[mobHash] or mobHash
end

function ThreatLib:TranslateForeignMobName(mobName)
	if not self.mobNameTranslationEnabled then
		return mobName
	end
	if mobNameTranslations[mobName] and mobNameTranslations[mobName][MyLocale] then
		return mobNameTranslations[mobName][MyLocale]
	end
	return mobName
end

function ThreatLib:UNIT_TARGET(unit)
	if not self.mobNameTranslationEnabled then
		return
	end
	
	if unit == "player" then
		return
	end
	
	-- Make sure it isn't a pet
	if not UnitPlayerControlled(unit) then
		return
	end

	local targetUnit = unit .. "target"
	local UnitName_unit = UnitName(unit)
	local foreignLocale = partyMemberLocales[UnitName_unit]
	
	if UnitIsUnit("target", targetUnit) and not UnitPlayerControlled(targetUnit) and foreignLocale ~= MyLocale then
		local localMobName = UnitName("target")
		-- We already have a translation for the specified locale/mob
		if mobNameTranslations[localMobName] and mobNameTranslations[localMobName][foreignLocale] then
			return
		end

		-- We don't have it, ask them for the name
		self:SendCommMessage("GROUP", "MOB_NAME_TRANSLATION_REQUEST", UnitName_unit, localMobName, MyLocale)
	end
end

------------------------------------------------
-- Debugging methods
------------------------------------------------

function ThreatLib:DumpThreatTable(sender)
	if sender then
		self:PrintLiteral(threatTargets[sender])
	else
		self:PrintLiteral(threatTargets)
	end
end

function ThreatLib:DumpTranslationTable(mobName)
	if mobName then
		self:PrintLiteral(mobNameTranslations[mobName])
	else
		self:PrintLiteral(mobNameTranslations)
	end
end

function ThreatLib:DumpTranslationHashes(mobName)
	if mobName then
		self:PrintLiteral(translationHashes[mobName])
	else
		self:PrintLiteral(translationHashes)
	end
end

ThreatLib.UpvalueFixers = {}
tinsert(ThreatLib.UpvalueFixers, function(lib)
	ThreatLib = lib
	local _
	_, MINOR_VERSION = lib:GetLibraryVersion()
	
	-- internal variables
	ThreatLib.new = nil
	ThreatLib.newHash = nil
	ThreatLib.newSet = nil
	ThreatLib.del = nil
end)
